from sqlalchemy import func, select
from infosystem.common.subsystem import driver


class Driver(driver.Driver):

    def get_nextval(self, session, domain_id, name):
        result = None
        get_nextval_func = func.domain_seq_nextval
        statement = select(get_nextval_func(domain_id, name))
        row_tuple = session.execute(statement).first()
        if len(row_tuple) > 0:
            result = row_tuple[0]
        return result



from infosystem.common import exception
from infosystem.subsystem.domain import tasks
from infosystem.common.subsystem import operation, manager
from infosystem.subsystem.image.resource import QualityImage


class DomainByName(operation.Operation):

    def pre(self, session, domain_name, **kwargs):
        self.domain_name = domain_name
        return True

    def do(self, session, **kwargs):
        domains = self.manager.list(name=self.domain_name)
        if not domains:
            raise exception.NotFound('ERROR! Domain name not found')
        domain = domains[0]

        # Hide user ID and settings from public resources
        domain.created_by = None
        domain.updated_by = None

        return domain


class DomainLogoByName(operation.Operation):

    def pre(self, session, domain_name, **kwargs):
        self.domain_name = domain_name
        return True

    def do(self, session, **kwargs):
        domains = self.manager.list(name=self.domain_name)
        if not domains:
            raise exception.NotFound('ERROR! Domain name not found')
        domain = domains[0]

        if domain.logo_id is None:
            raise exception.NotFound('ERROR! Domain logo not found')

        kwargs['quality'] = kwargs.get('quality', QualityImage.med)
        return self.manager.api.images().get(id=domain.logo_id, **kwargs)


class UploadLogo(operation.Update):

    def pre(self, session, id, token, file, **kwargs):
        self.file = file
        self.token = token

        return super().pre(session, id, **kwargs)

    def do(self, session, **kwargs):
        kwargs = {}
        kwargs['domain_id'] = self.entity.id
        kwargs['user_id'] = self.token.user_id
        kwargs['type_image'] = 'DomainLogo'

        image = self.manager.api.images().create(file=self.file, **kwargs)

        self.entity.logo_id = image.id

        return super().do(session=session)


class RemoveLogo(operation.Update):

    def do(self, session, **kwargs):
        logo_id = self.entity.logo_id
        self.entity.logo_id = None
        entity = super().do(session=session)
        if logo_id and entity is not None:
            self.manager.api.images().delete(id=logo_id)

        return


class Register(operation.Create):

    def _register_domain(self, domain_name, domain_display_name,
                         application_id, username, email, password):
        self.domain = self.manager.api.domains().create(
            application_id=application_id, name=domain_name,
            display_name=domain_display_name,
            addresses=[], contacts=[], active=False)

        self.user = self.manager.api.users().create(
            name=username, email=email, domain_id=self.domain.id, active=False)

        self.manager.api.users().reset(id=self.user.id, password=password)

    def pre(self, session, username, email, password,
            domain_name, domain_display_name, application_name):
        self.username = username
        self.email = email
        self.password = password
        self.domain_name = domain_name
        self.domain_display_name = domain_display_name
        self.application_name = application_name

        user_params_ok = username and email and password
        app_params_ok = domain_name and application_name
        if not user_params_ok and app_params_ok:
            raise exception.BadRequest(
                'ERROR! Not enough data to register domain')

        applications = \
            self.manager.api.applications().list(name=application_name)
        if not applications:
            raise exception.BadRequest('ERROR! Application name not found.')
        self.application = applications[0]

        return True

    def do(self, session, **kwargs):
        domains = self.manager.api.domains().list(name=self.domain_name)
        if not domains:
            self._register_domain(
                self.domain_name, self.domain_display_name, self.application.id,
                self.username, self.email, self.password)
        else:
            domain = domains[0]

            users = self.manager.api.users().list(email=self.email,
                                                  domain_id=domain.id)

            if domain.active:
                raise exception.BadRequest('Domain already activated')

            if not users or domain.display_name != self.domain_display_name:
                raise exception.BadRequest('Domain already registered')

            self.user = users[0]
            self.manager.api.users().reset(id=self.user.id,
                                           password=self.password)

        return True

    def post(self):
        # The notification don't be part of transaction must be on post
        tasks.send_email(self.user.id)


class Activate(operation.Create):

    def pre(self, session, token_id, domain_id, user_admin_id):
        if not (user_admin_id or domain_id):
            raise exception.BadRequest(
                'ERROR! Not enough data to Activate Domain')

        self.token_id = token_id
        self.domain_id = domain_id
        self.user_admin_id = user_admin_id

        roles = self.manager.api.roles().list(name='Admin')
        if not roles:
            raise exception.BadRequest('ERROR! Role Admin not found')
        self.role_admin = roles[0]

        return True

    def do(self, session, **kwargs):
        self.manager.api.domains().update(id=self.domain_id, active=True)
        self.manager.api.users().update(id=self.user_admin_id, active=True)
        self.manager.api.grants().create(user_id=self.user_admin_id,
                                         role_id=self.role_admin.id)
        self.manager.api.tokens().delete(id=self.token_id)

        domain = self.manager.api.domains().get(id=self.domain_id)
        if not domain:
            raise exception.BadRequest('ERROR! Domain not found')

        return domain


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetDomainSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class Manager(manager.Manager):

    def __init__(self, driver):
        super(Manager, self).__init__(driver)
        self.domain_by_name = DomainByName(self)
        self.domain_logo_by_name = DomainLogoByName(self)
        self.upload_logo = UploadLogo(self)
        self.remove_logo = RemoveLogo(self)
        self.register = Register(self)
        self.activate = Activate(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_domain_settings_by_keys = GetDomainSettingsByKeys(self)



from infosystem.database import db
from sqlalchemy import UniqueConstraint
from infosystem.common.subsystem import entity


class Tag(entity.Entity, db.Model):

    attributes = ['domain_id', 'name', 'color', 'description']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    name = db.Column(db.String(60), nullable=False)
    color = db.Column(db.CHAR(7), nullable=False)
    description = db.Column(db.String(1024), nullable=False)

    __table_args__ = (
        UniqueConstraint('domain_id', 'name', name='tag_domain_id_name_uk'),)

    def __init__(self, id, domain_id, name, color, description,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.name = name
        self.color = color
        self.description = description



from typing import Any, Callable, List, Optional, Type

import flask
from infosystem.common.subsystem.controller import Controller
from infosystem.common.subsystem.driver import Driver
from infosystem.common.subsystem.entity import Entity
from infosystem.common.subsystem.manager import Manager
from infosystem.common.subsystem.router import Router
from infosystem.common.subsystem.transaction_manager import TransactionManager


class Subsystem(flask.Blueprint):

    def __init__(self, resource: Optional[Type[Entity]] = None,
                 router: Optional[Type[Router]] = None,
                 controller: Optional[Type[Controller]] = None,
                 manager: Optional[Type[Manager]] = None,
                 driver: Optional[Type[Driver]] = None,
                 individual_name: Optional[str] = None,
                 collection_name: Optional[str] = None,
                 operations: List[Any] = []) -> None:

        self.__individual_name: str = individual_name or resource.individual()
        self.__collection_name: str = collection_name or resource.collection()

        super().__init__(name=self.__collection_name,
                         import_name=self.__collection_name)

        self.name = self.__collection_name
        self.__resource = resource
        self.__driver = driver if driver else Driver
        self.__controller = controller if controller else Controller
        self.__manager = manager if manager else Manager
        self.__router = router if router else Router
        self.router = self.__router(self.__collection_name, routes=operations)

        self.register_routes(self.router.routes)

        self.api = None

    def __validate_routes(self, routes: List[Router],
                          controller: Type[Controller]) -> List[str]:
        errors: List[str] = []
        for route in routes:
            callback_str = route['callback']

            if not hasattr(controller, callback_str):
                message = '{name} controller has no function {function}'.\
                    format(name=self.name, function=callback_str)
                errors.append(message)
                return errors

            fn = getattr(controller, callback_str)
            if not callable(fn):
                message = '{x} in {name} controller is not a function'.\
                    format(name=self.name, x=callback_str)
                errors.append(message)
        return errors

    def __view_func(self, callback_str: str) -> Callable:
        def wrapper(*args, **kwargs):
            transaction_manager = TransactionManager()
            controller = self.controller(transaction_manager)

            if not hasattr(controller, callback_str):
                message = '{name} controller has no function {function}'.\
                    format(name=self.name, function=callback_str)
                raise Exception(message)

            fn = getattr(controller, callback_str)
            if not callable(fn):
                message = '{x} in {name} controller is not a function'.\
                    format(name=self.name, x=callback_str)
                raise Exception(message)

            result =  fn(*args, **kwargs)
            try:
                transaction_manager.shutdown()
            except Exception as e:
                pass
            return result
        wrapper.__name__ = callback_str

        return wrapper

    def register_routes(self, routes: List[Router]) -> None:
        for route in self.router.routes:
            self.add_url_rule(rule=route['url'],
                              view_func=self.__view_func(route['callback']),
                              methods=[route['method']])

    def validate_routes(self) -> List[str]:
        return self.__validate_routes(self.router.routes, self.__controller)

    def lazy_manager(self, transaction_manager: TransactionManager) -> Callable[[TransactionManager], Manager]:
        return self.manager(transaction_manager)

    def manager(self, transaction_manager: TransactionManager) -> Manager:
        def instantiate(transaction_manager: TransactionManager) -> Manager:
            driver = self.driver(transaction_manager)
            manager = self.__manager(driver)
            return manager

        return instantiate(transaction_manager)

    def controller(self, transaction_manager: TransactionManager) -> Controller:
        def instantiate(transaction_manager: TransactionManager) -> Controller:
            if self.api is not None:
                api = self.api(transaction_manager)
                manager_cls = getattr(api, self.name)
                manager = manager_cls()
            else:
                manager = self.manager(transaction_manager)
            return self.__controller(manager,
                                     self.__individual_name,
                                     self.__collection_name)

        return instantiate(transaction_manager)

    def driver(self, transaction_manager: TransactionManager) -> Optional[Driver]:
        def instantiate(transaction_manager: TransactionManager) -> Optional[Driver]:
            return self.__driver(self.__resource, transaction_manager) if self.__resource else None

        return instantiate(transaction_manager)


from typing import Dict, List

from infosystem.common.subsystem import Subsystem
from infosystem.common.subsystem.apihandler import Api
from infosystem.common.input import RouteResource
from infosystem.subsystem.role.resource import Role
from infosystem.bootstrap.roles import BootstrapRoles
from infosystem.bootstrap.default.application import BootstrapApplication
from infosystem.bootstrap.default.domain import BootstrapDomain
from infosystem.bootstrap.default.user import BootstrapUser
from infosystem.bootstrap.default.policies import BootstrapPolicies


class BootstrapDefault(object):

    def __init__(self, api: Api) -> None:
        self.bootstrap_roles = BootstrapRoles(api)
        self.bootstrap_application = BootstrapApplication(api)
        self.bootstrap_domain = BootstrapDomain(api)
        self.bootstrap_user = BootstrapUser(api)
        self.bootstrap_policies = BootstrapPolicies(api)

    def execute(self, user_resources: List[RouteResource],
                sysadmin_resources: List[RouteResource],
                sysadmin_exclusive_resources: List[RouteResource]):
        roles = self.bootstrap_roles.execute()
        role_sysadmin = self._get_role_sysadmin(roles)

        application = self.bootstrap_application.execute()
        domain = self.bootstrap_domain.execute(application.id)
        self.bootstrap_user.execute(domain.id, role_sysadmin.id)
        self.bootstrap_policies.execute(application.id,
                                        role_sysadmin.id,
                                        user_resources,
                                        sysadmin_resources,
                                        sysadmin_exclusive_resources)

    def _get_role_sysadmin(self, roles: List[Role]) -> Role:
        role = next((role for role in roles if role.name == Role.SYSADMIN),
                    None)
        if not role:
            raise Exception()
        return role



from flask import Flask
from flask.globals import current_app
from celery import Celery

celery = Celery('infosystem')


def init_celery(app: Flask) -> Celery:
    celery.conf.update(broker_url=app.config['CELERY_BROKER_URL'])
    #    result_backend=app.config['CELERY_BACKEND_URL'])

    class ContextTask(celery.Task):
        def __call__(self, *args, **kwargs):
            with app.app_context():
                return self.run(*args, **kwargs)

    celery.Task = ContextTask
    return celery


def decide_on_run(fn):
    def wrapper(*args):
        should_use_worker = current_app.config['USE_WORKER']
        if should_use_worker:
            return fn.delay(*args)
        else:
            return fn(*args)
    return wrapper



from sqlalchemy import orm
from infosystem.database import db
from infosystem.common.subsystem import entity


class TimelineEvent(entity.Entity, db.Model):

    LIMIT_SEARCH = 30

    attributes = ['domain_id', 'event_at', 'event_by', 'lat', 'lon',
                  'description', 'entity', 'entity_id']
    attributes += entity.Entity.attributes

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    event_at = db.Column(db.DateTime, nullable=False, unique=False)
    event_by = db.Column(db.CHAR(32), nullable=False, unique=False)
    lat = db.Column(db.Numeric(14, 8), nullable=False, unique=False)
    lon = db.Column(db.Numeric(14, 8), nullable=False, unique=False)
    description = db.Column(db.String(500), nullable=False, unique=False)
    entity = db.Column(db.String(100), nullable=True, unique=False)
    entity_id = db.Column(db.CHAR(32), nullable=True, unique=False)
    users = orm.relationship(
        "TimelineEventUser", backref=orm.backref('timeline_event_user'),
        cascade='delete,delete-orphan,save-update')

    __tablename__ = 'timeline_event'

    def __init__(self, id, domain_id, event_at, event_by, lat, lon,
                 description, entity=None, entity_id=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.id = id
        self.domain_id = domain_id
        self.event_at = event_at
        self.event_by = event_by
        self.lat = lat
        self.lon = lon
        self.description = description
        self.entity = entity
        self.entity_id = entity_id,

    @classmethod
    def individual(cls):
        return 'timeline_event'

    @classmethod
    def embedded(cls):
        return ['users']


class TimelineEventUser(entity.Entity, db.Model):
    attributes = ['id', 'user_id']

    timeline_event_id = db.Column(
        db.CHAR(32), db.ForeignKey("timeline_event.id"), nullable=False)
    user_id = db.Column(
        db.CHAR(32), db.ForeignKey("user.id"), nullable=False)
    user = orm.relationship(
        'User', backref=orm.backref('timeline_event_user'))

    def __init__(self, id, timeline_event_id, user_id,
                 active=True, created_at=None,
                 created_by=None, updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.timeline_event_id = timeline_event_id
        self.user_id = user_id

    def is_stable(self):
        if self.user_id is not None and self.timeline_event_id is not None:
            return True
        return False



from infosystem.common import utils
from infosystem.common.subsystem.apihandler import Api
from infosystem.subsystem.grant.resource import Grant
from infosystem.subsystem.user.resource import User


class BootstrapUser(object):

    def __init__(self, api: Api) -> None:
        self.user_manager = api.users()
        self.grant_manager = api.grants()

    def execute(self, domain_id: str, role_id: str) -> User:
        user = self._get_user_default(domain_id)
        grant = self._get_grant_default(user.id, role_id)
        self._save(user, grant)
        return user

    def _get_user_default(self, domain_default_id: str) -> User:
        return User(id=utils.random_uuid(),
                    domain_id=domain_default_id,
                    name=User.SYSADMIN_USERNAME,
                    email='sysadmin@example.com')

    def _get_grant_default(self, user_id: str, role_id: str) -> Grant:
        return Grant(id=utils.random_uuid(),
                     user_id=user_id,
                     role_id=role_id)

    def _save(self, user: User, grant: Grant) -> None:
        self.user_manager.create(**user.to_dict())
        self.user_manager.reset(id=user.id, password='123456')
        self.grant_manager.create(**grant.to_dict())



from infosystem.database import db
from sqlalchemy import UniqueConstraint, orm
from infosystem.common.subsystem import entity


class Capability(entity.Entity, db.Model):

    attributes = ['route_id', 'application_id']
    attributes += entity.Entity.attributes

    route_id = db.Column(
        db.CHAR(32), db.ForeignKey("route.id"), nullable=False)
    route = orm.relationship('Route', backref=orm.backref('capabilities'))
    application_id = db.Column(
        db.CHAR(32), db.ForeignKey("application.id"), nullable=False)
    application = orm.relationship('Application', backref=orm.backref(
        'capabilities'))

    __table_args__ = (
        UniqueConstraint('route_id', 'application_id', name='capability_uk'),)

    def __init__(self, id, route_id, application_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.route_id = route_id
        self.application_id = application_id

    @classmethod
    def collection(cls):
        return 'capabilities'



from infosystem.common import subsystem
from infosystem.subsystem.file import resource
from infosystem.subsystem.file import manager
from infosystem.subsystem.file import controller

subsystem = subsystem.Subsystem(resource=resource.File,
                                manager=manager.Manager,
                                controller=controller.Controller)



from infosystem.database import db
from infosystem.common.subsystem import entity
from sqlalchemy import orm


class Grant(entity.Entity, db.Model):

    attributes = ['user_id', 'role_id']
    attributes += entity.Entity.attributes

    user_id = db.Column(db.CHAR(32), db.ForeignKey("user.id"), nullable=False)
    user = orm.relationship("User", backref=orm.backref('grants'))
    role_id = db.Column(db.CHAR(32), db.ForeignKey("role.id"), nullable=False)
    role = orm.relationship("Role", backref=orm.backref('grants'))

    def __init__(self, id, user_id, role_id,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.user_id = user_id
        self.role_id = role_id



from infosystem.database import db
from sqlalchemy import UniqueConstraint
from infosystem.common.subsystem import entity


class Role(entity.Entity, db.Model):

    USER = 'User'
    SYSADMIN = 'Sysadmin'
    ADMIN = 'Admin'

    attributes = ['name']
    attributes += entity.Entity.attributes

    name = db.Column(db.String(80), nullable=False)

    __table_args__ = (
        UniqueConstraint('name', name='role_name_uk'),)

    def __init__(self, id, name,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.name = name



import json
import uuid
import unittest
import testtools

from infosystem.tests.functional import test_base


class PolicyTestCase(test_base.CRUDTest): #, testtools.TestCase):

    def load_fixtures(self):
        domain = self.post(
            resource_ref={'name': uuid.uuid4().hex, 'active': True},
            resource_name='domain',
            collection_name='domains')

        self.domain_id = domain['id']

    @property
    def resource_name(self):
        return 'policy'

    @property
    def required_attributes(self):
        return ['domain_id', 'name', 'url', 'method']

    @property
    def unique_attributes(self):
        return [('domain_id', 'name', 'url', 'method')]

    def new_resource_ref(self):
        return {'domain_id': self.domain_id,
                'name': uuid.uuid4().hex,
                'url': '/test',
                'method': 'GET'}


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from decimal import Decimal
from datetime import datetime

from infosystem.database import db

DATE_FMT = '%Y-%m-%d'
DATETIME_FMT = '%Y-%m-%dT%H:%M:%S.%fZ'


class Entity(object):

    attributes = ['id', 'active', 'created_at', 'created_by',
                  'updated_at', 'updated_by', 'tag']
    pagination_column = 'id desc'

    id = db.Column(db.CHAR(32), primary_key=True, autoincrement=False)
    active = db.Column(db.Boolean())
    created_at = db.Column(db.DateTime)
    created_by = db.Column(db.CHAR(32))
    updated_at = db.Column(db.DateTime)
    updated_by = db.Column(db.CHAR(32))
    tag = db.Column(db.String(1000))

    def __init__(self, id, active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        self.id = id
        self.active = active
        self.created_at = created_at
        self.created_by = created_by
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.tag = tag

    @classmethod
    def embedded(cls):
        return []

    @classmethod
    def individual(cls):
        return cls.__name__.lower()

    @classmethod
    def collection(cls):
        return cls.individual() + 's'

    def is_stable(self):
        return True

    def allDateFmtFromAllTypes(self, dateOrDateTime):
        dateTime = None
        if dateOrDateTime is not None:
            if type(dateOrDateTime) is str:
                try:
                    if len(dateOrDateTime.strip()) == 10:
                        dateTime = datetime.strptime(
                            dateOrDateTime, DATE_FMT)
                    elif len(dateOrDateTime.strip()) == 24:
                        dateTime = datetime.strptime(
                            dateOrDateTime, DATETIME_FMT)
                except Exception:
                    pass
            else:
                if type(dateOrDateTime) is datetime:
                    dateTime = dateOrDateTime

        return dateTime

    def convert_numeric(self, value):
        if str(value).find('.') > -1:
            return float(value.real)
        else:
            return int(value)

    def to_dict(self, include_dict=None, stringify=True):
        d = {}
        include_dict = include_dict or {}

        for attr in self.__class__.attributes:
            if attr not in include_dict:
                value = getattr(self, attr)
                if value is not None:
                    if isinstance(value, Enum):
                        d[attr] = value.name
                    elif isinstance(value, Decimal):
                        d[attr] = self.convert_numeric(value)
                    elif isinstance(value, Entity):
                        include_dict.update({attr: {}})
                    else:
                        d[attr] = value
                # TODO(fdoliveira) Why change format of date and datetime?
                # if stringify and isinstance(value, datetime):
                #    d[attr] = value.strftime(DATETIME_FMT)
                # elif stringify and isinstance(value, date):
                #    d[attr] = value.strftime(DATETIME_FMT)
                # else:

        # Only embedded that are not in include will be updated
        # include_dict.update({attr: {} for attr in self.embedded()})
        for attr in self.embedded():
            if attr not in include_dict:
                include_dict.update({attr: {}})

        if include_dict:
            for key, value in include_dict.items():
                if not isinstance(value, dict):
                    # it's a filter
                    if getattr(self, key) != value:
                        raise AssertionError()
                    continue

                thing = getattr(self, key)
                if isinstance(thing, list):
                    values = []
                    empty = True
                    for part in thing:
                        try:
                            values.append(part.to_dict(value))
                            empty = False
                        except AssertionError:
                            # filter mismatch, ignore the expansion
                            pass
                    if values and empty:
                        # filter mismatch, no entity matched the filter,
                        # re-raise and ignore current entity
                        raise AssertionError()
                    d[key] = [part.to_dict(value) for part in thing]
                else:
                    try:
                        if thing is not None:
                            d[key] = thing.to_dict(value)
                    except AssertionError:
                        # filter mismatch, re-raise to ignore current entity
                        raise

        return d

    def get_last_user_id_who_modified(self):
        if self.updated_by is not None:
            return self.updated_by
        else:
            return self.created_by



from infosystem import database


class TransactionManager(object):

    def __init__(self, session=None) -> None:
        self.session = database.db.session
        self.count = 0

    def begin(self):
        self.count += 1

    def commit(self):
        self.count -= 1
        if self.count == 0:
            self.session.commit()

    def rollback(self):
        self.session.rollback()
        self.count = -1000000

    def shutdown(self):
        pass


import flask
from infosystem.common.subsystem import controller
from infosystem.common import exception
from infosystem.subsystem.user.email import TypeEmail


class Controller(controller.Controller):

    def _get_token_id(self):
        return flask.request.headers.get('token')

    def _get_application_by_name(self, name):
        applications = self.manager.api.applications().list(name=name)
        if not applications:
            raise exception.BadRequest()
        return applications[0]

    def _get_register_data(self):
        data = flask.request.get_json()

        username = data.get('username', 'admin')
        email = data.get('email', None)
        password = data.get('password', None)
        domain_name = data.get('domain', None)
        application_name = data.get('application', None)

        if not (username and email and password and domain_name and
                application_name):
            raise exception.BadRequest()

        return (username, email, password, domain_name, application_name)

    def register(self):

        try:
            (username, email, password, domain_name, application_name) = \
                self._get_register_data()

            application = self._get_application_by_name(application_name)

            domain = self.manager.api.domains().create(
                name=domain_name, application_id=application.id, active=False)
            user = self.manager.api.users().create(
                name=username, email=email, domain_id=domain.id, active=False)
            self.manager.api.users().reset(id=user.id, password=password)
            self.manager.api.users().notify(
                id=user.id, type_email=TypeEmail.ACTIVATE_ACCOUNT)

        except exception.InfoSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")

    def _get_activate_data(self):
        data = flask.request.get_json()

        user_id = data.get('user_id', None)
        domain_id = data.get('domain_id', None)

        if not (user_id and domain_id):
            raise exception.BadRequest()

        return (user_id, domain_id)

    def _get_role_admin(self):
        roles = self.manager.api.roles().list(name='Admin')
        if not roles:
            raise exception.BadRequest()
        return roles[0]

    def activate(self):
        try:
            (user_id, domain_id) = self._get_activate_data()

            domain = self.manager.api.domains().get(id=domain_id)
            user = self.manager.api.users().get(id=user_id)

            if not (domain and user):
                raise exception.BadRequest()
            role_admin = self._get_role_admin()

            self.manager.api.domains().update(id=domain.id, active=True)
            self.manager.api.users().update(id=user.id, active=True)
            self.manager.api.grants().create(user_id=user.id,
                                             role_id=role_admin.id)
            self.manager.api.tokens().delete(id=self._get_token_id())
        except exception.InfoSystemException as exc:
            return flask.Response(response=exc.message,
                                  status=exc.status)

        return flask.Response(response=None,
                              status=204,
                              mimetype="application/json")



from gabbi import fixture

import abc
import json
import uuid
import infosystem
import os


class Fixture(fixture.GabbiFixture):

    def __init__(self):
        super().__init__()

        self.system = infosystem.SystemFlask()
        self.app = self.system.test_client()

    @property
    def individual(self):
        return None

    @property
    def collection(self):
        return self.individual + 's'

    @abc.abstractmethod
    def new_dict(self):
        raise NotImplementedError

    def start_fixture(self):
        if self.individual == 'token':
            headers = {'Content-Type': 'application/json'}
        else:
            headers = {'token': self.env('token'),
                       'Content-Type': 'application/json'}

        response = self.app.post(
            '/' + self.collection,
            headers=headers,
            data=json.dumps(self.new_dict()))

        resource = json.loads(response.data.decode())[self.individual]

        k = self.individual
        if self.env(k):
            i = 2
            while self.env(k + '.' + str(i)):
                i += 1
            k = k + '.' + str(i)

        self.key = k
        self.value = resource['id']

        self.env(self.key, self.value)

    def stop_fixture(self):
        if self.collection == 'tokens':
            return
        headers = {'token': self.env('token'),
                   'Content-Type': 'application/json'}
        self.app.delete('/' + self.collection + '/' + self.value,
                        headers=headers)
        os.environ.pop(self.key)

    def env(self, k, v=None):
        if v is not None:
            os.environ[k] = v
        else:
            try:
                v = os.environ[k]
            except KeyError:
                v = None
        return v


class TokenFixture(Fixture):

    @property
    def individual(self):
        return 'token'

    def new_dict(self):
        return {'domain_name': 'default',
                'username': 'admin',
                'password': '123456'}


class DomainFixture(Fixture):

    @property
    def individual(self):
        return 'domain'

    def new_dict(self):
        return {'name': uuid.uuid4().hex}


class UserFixture(Fixture):

    @property
    def individual(self):
        return 'user'

    def new_dict(self):
        return {'domain_id': self.env('domain'),
                'name': uuid.uuid4().hex,
                'email': uuid.uuid4().hex,
                'password': uuid.uuid4().hex}



import json
from typing import Any
from sqlalchemy import orm

from infosystem.database import db
from infosystem.common.subsystem import entity
from infosystem.common import exception


class Domain(entity.Entity, db.Model):

    DEFAULT = 'default'

    attributes = ['name', 'display_name', 'parent_id',
                  'application_id', 'logo_id', 'doc', 'description', 'settings']
    attributes += entity.Entity.attributes

    application_id = db.Column(
        db.CHAR(32), db.ForeignKey("application.id"), nullable=False)
    application = orm.relationship('Application', backref=orm.backref(
        'domains'))
    name = db.Column(db.String(60), nullable=False, unique=True)
    display_name = db.Column(db.String(100), nullable=False)
    doc = db.Column(db.String(60), nullable=True)
    description = db.Column(db.String(1000), nullable=True)
    logo_id = db.Column(db.CHAR(32), db.ForeignKey('image.id'), nullable=True)
    parent_id = db.Column(
        db.CHAR(32), db.ForeignKey("domain.id"), nullable=True)
    addresses = orm.relationship(
        'DomainAddress', backref=orm.backref('domain_addresses'),
        cascade='delete,delete-orphan,save-update')
    contacts = orm.relationship(
        'DomainContact', backref=orm.backref('domain_contacts'),
        cascade='delete,delete-orphan,save-update')
    _settings = db.Column('settings', db.Text, nullable=False, default='{}')

    __tablename__ = 'domain'

    def __init__(self, id, application_id, name, display_name=None,
                 doc=None, description=None, logo_id=None, parent_id=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.application_id = application_id
        self.name = name
        if display_name is None:
            self.display_name = name
        else:
            self.display_name = display_name
        self.doc = doc
        self.description = description
        self.logo_id = logo_id
        self.parent_id = parent_id

    def _has_setting(self, key: str) -> bool:
        return self.settings.get(key) is not None

    def remove_setting(self, key: str):
        if not self._has_setting(key):
            raise exception.BadRequest(f"Erro! Setting {key} not exists")

        settings = self.settings
        value = settings.pop(key)
        self._save_settings(settings)

        return value

    def update_setting(self, key: str, value: Any):
        settings = self.settings
        settings[key] = value
        self._save_settings(settings)
        return value

    @property
    def settings(self):
        try:
            settings_str = '{}' if self._settings is None else self._settings
            return json.loads(settings_str)
        except Exception:
            return {}

    def _save_settings(self, settings: dict):
        self._settings = json.dumps(settings, default=str)

    @classmethod
    def embedded(cls):
        return ['addresses', 'contacts']


class DomainAddress(entity.Entity, db.Model):

    attributes = ['postal_code', 'address_line_1', 'address_line_2',
                  'city', 'state_province', 'country', 'tag']

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey('domain.id'), nullable=False)
    postal_code = db.Column(db.String(30), nullable=False)
    address_line_1 = db.Column(db.String(255), nullable=False)
    address_line_2 = db.Column(db.String(255), nullable=True)
    city = db.Column(db.String(100), nullable=False)
    state_province = db.Column(db.String(60), nullable=False)
    country = db.Column(db.String(60), nullable=False)

    def __init__(self, id, domain_id, postal_code, address_line_1,
                 city, state_province, country, address_line_2=None,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.postal_code = postal_code
        self.address_line_1 = address_line_1
        self.address_line_2 = address_line_2
        self.city = city
        self.state_province = state_province
        self.country = country


class DomainContact(entity.Entity, db.Model):

    attributes = ['contact', 'tag']

    domain_id = db.Column(
        db.CHAR(32), db.ForeignKey("domain.id"), nullable=False)
    contact = db.Column(db.String(100), nullable=False)

    def __init__(self, id, domain_id, contact,
                 active=True, created_at=None, created_by=None,
                 updated_at=None, updated_by=None, tag=None):
        super().__init__(id, active, created_at, created_by,
                         updated_at, updated_by, tag)
        self.domain_id = domain_id
        self.contact = contact



from infosystem.common.subsystem.apihandler import Api
import uuid

from typing import Dict, List

from infosystem.common.subsystem import Subsystem
from infosystem.subsystem.role.resource import Role


class BootstrapRoles(object):

    def __init__(self, api: Api) -> None:
        self.role_manager = api.roles()

    def execute(self):
        roles = self.role_manager.list()
        if not roles:
            default_roles = self._default_roles()
            roles = self.role_manager.create_roles(roles=default_roles)
        return roles

    def _get_role(self, name: str) -> Role:
        role = Role(id=uuid.uuid4().hex, name=name)
        return role

    def _default_roles(self) -> List[Role]:
        user = self._get_role(Role.USER)
        sysadmin = self._get_role(Role.SYSADMIN)
        admin = self._get_role(Role.ADMIN)

        return [user, sysadmin, admin]



from infosystem.common import subsystem
from infosystem.subsystem.registration import controller, router

subsystem = subsystem.Subsystem(
    individual_name='register',
    collection_name='register',
    controller=controller.Controller,
    router=router.Router)



from infosystem.common.subsystem.driver import Driver
from infosystem.common.subsystem import operation


class Manager(object):

    def __init__(self, driver: Driver) -> None:
        self.driver = driver

        self.create = operation.Create(self)
        self.get = operation.Get(self)
        self.list = operation.List(self)
        self.update = operation.Update(self)
        self.delete = operation.Delete(self)
        # NOTE(samueldmq): what do we use this for ?
        self.count = operation.Count(self)



from infosystem.common.subsystem.apihandler import Api
from typing import Dict

from infosystem.common import utils
from infosystem.common.subsystem import Subsystem
from infosystem.subsystem.domain.resource import Domain


class BootstrapDomain(object):

    def __init__(self, api: Api) -> None:
        self.domain_manager = api.domains()

    def execute(self, application_id: str) -> Domain:
        domain = self._get_domain_default(application_id)
        return self._save_domain(domain)

    def _get_domain_default(self, application_default_id: str) -> Domain:
        return Domain(id=utils.random_uuid(),
                      application_id=application_default_id,
                      name=Domain.DEFAULT)

    def _save_domain(self, domain: Domain) -> Domain:
        data = domain.to_dict()
        if 'settings' in data.keys():
            data.pop('settings')
        data['addresses'] = []
        data['contacts'] = []
        return self.domain_manager.create(**data)



from infosystem.common.subsystem import router


class Router(router.Router):

    def __init__(self, collection, routes=[]):
        super().__init__(collection, routes)

    @property
    def routes(self):
        return [
            {
                'action': 'create',
                'method': 'POST',
                'url': self.collection_url,
                'callback': 'create'
            },
            {
                'action': 'get',
                'method': 'GET',
                'bypass': True,
                'url': self.resource_url,
                'callback': 'get'
            },
            {
                'action': 'list',
                'method': 'GET',
                'url': self.collection_url,
                'callback': 'list'
            },
            {
                'action': 'update',
                'method': 'PUT',
                'url': self.resource_url,
                'callback': 'update'
            },
            {
                'action': 'delete',
                'method': 'DELETE',
                'url': self.resource_url,
                'callback': 'delete'
            }
        ]



from infosystem.common import subsystem
from infosystem.subsystem.domain import manager, resource, controller, router

subsystem = subsystem.Subsystem(resource=resource.Domain,
                                controller=controller.Controller,
                                manager=manager.Manager,
                                router=router.Router)



from infosystem.common import subsystem
from infosystem.subsystem.user import resource

from infosystem.subsystem.user import controller
from infosystem.subsystem.user import manager
from infosystem.subsystem.user import router


subsystem = subsystem.Subsystem(resource=resource.User,
                                router=router.Router,
                                controller=controller.Controller,
                                manager=manager.Manager)



from sqlalchemy import and_
from infosystem.subsystem.application.resource import Application
from typing import List
from infosystem.common import exception
from infosystem.common.input import RouteResource, InputResource, \
    InputResourceUtils
from infosystem.common.subsystem import operation, manager
from infosystem.subsystem.capability.resource import Capability
from infosystem.subsystem.policy.resource import Policy
from infosystem.subsystem.role.resource import Role
from infosystem.subsystem.route.resource import Route


class Create(operation.Create):

    def pre(self, session, **kwargs) -> bool:
        self.exceptions = kwargs.pop('exceptions', [])

        return super().pre(session, **kwargs)

    def do(self, session, **kwargs):
        super().do(session)
        self.manager.create_user_capabilities_and_policies(id=self.entity.id,
                                                           session=session)
        if self.entity.name != Application.DEFAULT:
            self.manager.create_admin_capabilities_and_policies(
                id=self.entity.id, session=session, exceptions=self.exceptions)
        return self.entity


class CreateUserCapabilitiesAndPolicies(operation.Operation):

    def pre(self, session, id, **kwargs) -> bool:
        self.application_id = id
        self.user_resources = self.manager.bootstrap_resources.USER
        self.role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.USER).id

        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.user_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.role_id,
                                                 **self.resources)


class CreateAdminCapabilitiesAndPolicies(operation.Operation):

    def _map_routes(self, routes: List[Route]) -> List[RouteResource]:
        resources = [(route.url, route.method) for route in routes]
        return resources

    def _filter_resources(self, all_resources: List[RouteResource],
                          exceptions_resources: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions_resources)
        return resources

    def pre(self, session, id, exceptions: List[InputResource] = [], **kwargs):
        self.application_id = id
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)

        routes = self.manager.api.routes().list(active=True)
        routes_resources = self._map_routes(routes)

        self.admin_role_id = self.manager.api.roles().\
            get_role_by_name(role_name=Role.ADMIN).id

        self.admin_resources = self._filter_resources(
            routes_resources, exceptions_resources,
            self.manager.bootstrap_resources.SYSADMIN_EXCLUSIVE,
            self.manager.bootstrap_resources.USER)
        return True

    def do(self, session, **kwargs):
        self.resources = {'resources': self.admin_resources}
        self.manager.api.capabilities().create_capabilities(
            id=self.application_id, **self.resources)

        self.resources['application_id'] = self.application_id
        self.manager.api.roles().create_policies(id=self.admin_role_id,
                                                 **self.resources)


class CreateCapabilitiesWithExceptions(operation.Operation):

    def _filter_resources(self, routes: List[Route],
                          exceptions: List[RouteResource],
                          sysadmin_exclusive_resources: List[RouteResource],
                          user_resources: List[RouteResource]) \
            -> List[RouteResource]:
        all_resources = [(route.url, route.method) for route in routes]
        resources = InputResourceUtils.diff_resources(
            all_resources, sysadmin_exclusive_resources, user_resources,
            exceptions)
        return resources

    def pre(self, session, id: str, **kwargs):
        self.application_id = id
        exceptions = kwargs.get('exceptions', None)

        if not self.application_id or exceptions is None:
            raise exception.BadRequest()

        routes = self.manager.api.routes().list(active=True)
        exceptions_resources = InputResourceUtils.parse_resources(exceptions)
        self.resources = self.\
            _filter_resources(routes,
                              exceptions_resources,
                              self.manager.bootstrap_resoures.
                              SYSADMIN_EXCLUSIVE,
                              self.manager.bootstrap_resources.USER)

        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        data = {'resources', self.resources}
        self.manager.api.capabilities().\
            create_capabilities(id=self.application_id, **data)


class GetRoles(operation.Operation):

    def pre(self, session, id, **kwargs):
        self.application_id = id
        return self.driver.get(id, session=session) is not None

    def do(self, session, **kwargs):
        roles = session.query(Role). \
            join(Policy). \
            join(Capability). \
            filter(and_(Capability.application_id == self.application_id,
                        Role.name != Role.USER)). \
            distinct()
        return roles


class UpdateSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.settings = kwargs
        if self.settings is None or not self.settings:
            raise exception.BadRequest("Erro! There is not a setting")
        return super().pre(session=session, id=id)

    def do(self, session, **kwargs):
        result = {}
        for key, value in self.settings.items():
            new_value = self.entity.update_setting(key, value)
            result[key] = new_value
        super().do(session)

        return result


class RemoveSettings(operation.Update):

    def pre(self, session, id: str, **kwargs) -> bool:
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        super().pre(session, id=id)

        return self.entity.is_stable()

    def do(self, session, **kwargs):
        result = {}
        for key in self.keys:
            value = self.entity.remove_setting(key)
            result[key] = value
        super().do(session=session)

        return result


class GetApplicationSettingsByKeys(operation.Get):

    def pre(self, session, id, **kwargs):
        self.keys = kwargs.get('keys', [])
        if not self.keys:
            raise exception.BadRequest('Erro! keys are empty')
        return super().pre(session, id=id)

    def do(self, session, **kwargs):
        entity = super().do(session=session)
        settings = {}
        for key in self.keys:
            value = entity.settings.get(key, None)
            if value is not None:
                settings[key] = value
        return settings


class Manager(manager.Manager):

    def __init__(self, driver):
        super().__init__(driver)
        self.create = Create(self)
        self.create_user_capabilities_and_policies = \
            CreateUserCapabilitiesAndPolicies(self)
        self.create_admin_capabilities_and_policies = \
            CreateAdminCapabilitiesAndPolicies(self)
        self.get_roles = GetRoles(self)
        self.update_settings = UpdateSettings(self)
        self.remove_settings = RemoveSettings(self)
        self.get_application_settings_by_keys = \
            GetApplicationSettingsByKeys(self)
